import ipaddress
import random
import re
import string

_REGEX_IFNAME = r"[a-zA-Z][a-zA-Z0-9_]{1,14}"
_REGEX_NICKNAME = r"[a-zA-Z][a-zA-Z0-9_]{1,20}"


class ExitException(Exception):
    pass


class ArgsException(ExitException):
    pass


class ConfigVersionException(ExitException):
    pass


class InterfaceNotFoundException(ExitException):
    pass


class PeerNotFoundException(ExitException):
    pass


class Input:
    @staticmethod
    def check_int(
        value: str | int, min_value: int | None = None, max_value: int | None = None
    ):
        try:
            v = int(value)
            if min_value is not None and v < min_value:
                return False, f"Too small! (minimum is {min_value})"
            if max_value is not None and v > max_value:
                return False, f"Too large! (maximum is {max_value})"
        except ValueError:
            return False, "Value is not an integer."
        return True, ""

    @staticmethod
    def check_cidr4(value: str, optional: bool = False):
        if value:
            try:
                ipaddress.IPv4Network(value)
            except ValueError:
                return False, "Not a valid IPv4 CIDR block."
        else:
            if not optional:
                return False, "Value is required."
        return True, ""

    @staticmethod
    def check_cidr6(value: str, optional: bool = False):
        if value:
            try:
                ipaddress.IPv6Network(value)
            except ValueError:
                return False, "Not a valid IPv6 CIDR block."
        else:
            if not optional:
                return False, "Value is required."
        return True, ""

    @staticmethod
    def check_iface(value: str, optional: bool = False):
        if value:
            if not re.fullmatch(_REGEX_IFNAME, value):
                return False, "Not a valid interface name."
        else:
            if not optional:
                return False, "Value is required."
        return True, ""

    @staticmethod
    def check_peer_name(value: str, optional: bool = False):
        if value:
            if not re.fullmatch(_REGEX_NICKNAME, value):
                return (
                    False,
                    f"Not in {_REGEX_NICKNAME}",
                )  # TODO improve explanation / open up nickname regex
        else:
            if not optional:
                return False, "Value is required."
        return True, ""


class IP:
    @staticmethod
    def server_addr4(cidr4: str):
        mask = cidr4.split("/", 2)[1]
        return str(ipaddress.IPv4Network(cidr4)[1]) + f"/{mask}"

    @staticmethod
    def next_addr4(interface_cidr4: str, peers_cidr4: list[str]):
        network = ipaddress.IPv4Network(interface_cidr4)
        pools = [network]
        exclusions: list[ipaddress.IPv4Network] = [
            ipaddress.IPv4Network(network[0]),
            ipaddress.IPv4Network(network[1]),
        ]
        exclusions.extend(
            ipaddress.IPv4Network(peer_cidr4) for peer_cidr4 in peers_cidr4
        )
        for exclusion in exclusions:
            next_pools: list[ipaddress.IPv4Network] = []
            for pool in pools:
                try:
                    next_pools.extend(pool.address_exclude(exclusion))
                except ValueError:
                    # peer pool not in interface pool, skip it
                    next_pools.append(pool)
            pools = next_pools
        assert pools
        return f"{str(sorted(pools)[0][0])}/32"

    @staticmethod
    def auto_cidr4():
        """
        Returns a random /24 of private IPv4 addresses.
        Pool is chosen from 192.168.0.0/16, excluding 192.168.0.* to 192.168.10.*.
        These pools are excluded because they are commonly used for home
        networks and will likely conflict with the created VPN.
        """
        return f"192.168.{random.randint(11, 254)}.0/24"

    @staticmethod
    def server_addr6(cidr6: str):
        mask = cidr6.split("/", 2)[1]
        return str(ipaddress.IPv6Network(cidr6)[1]) + f"/{mask}"

    @staticmethod
    def next_addr6(interface_cidr6: str, peers_cidr6: list[str]):
        network = ipaddress.IPv6Network(interface_cidr6)
        pools = [network]
        exclusions: list[ipaddress.IPv6Network] = [
            ipaddress.IPv6Network(network[0]),
            ipaddress.IPv6Network(network[1]),
        ]
        exclusions.extend(
            ipaddress.IPv6Network(peer_cidr6) for peer_cidr6 in peers_cidr6
        )
        for exclusion in exclusions:
            next_pools: list[ipaddress.IPv6Network] = []
            for pool in pools:
                try:
                    next_pools.extend(pool.address_exclude(exclusion))
                except ValueError:
                    # peer pool not in interface pool, skip it
                    next_pools.append(pool)
            pools = next_pools
        assert pools
        return f"{str(sorted(pools)[0][0])}/128"

    @staticmethod
    def auto_cidr6():
        """
        Returns a random /64 of private IPv6 addresses.
        """
        rh = lambda x: "".join(
            random.choice(string.hexdigits) for i in range(x)
        ).lower()
        return f"fd{rh(2)}:{rh(4)}:{rh(4)}::/64"



import argparse
import logging
import sys
from enum import Enum
from typing import Any

from wgup import defaults, wireguard
from wgup.config import Config
from wgup.util import (
    IP,
    ArgsException,
    ExitException,
    Input,
    InterfaceNotFoundException,
    PeerNotFoundException,
)

_logger = logging.getLogger(defaults.PROG)
_logger.setLevel(logging.INFO)
_stderr_handler = logging.StreamHandler()
_stderr_handler.setLevel(logging.DEBUG)
_stderr_handler.setFormatter(logging.Formatter("{levelname:<8} : {message}", style="{"))
_logger.addHandler(_stderr_handler)

_FMT_INTERFACES = "{iface:15} : {host}:{port}"
_FMT_PEERS = "{name:20} : {cidr4:14} : {cidr6}"
_FMT_ATTRS = "{:20} : {}"


class Iface:
    class Attributes(Enum):
        NAME = "name"
        HOST = "host"
        PORT = "port"
        NAT_IFACE = "nat_iface"

    @staticmethod
    def _get(c: Config, args: argparse.Namespace):
        """
        This function intentionally uses args instead of the name of the
        interface to ensure that positional arguments are consistent across
        subcommands.
        """
        iface = c.interfaces.get(args.interface)
        if iface is None:
            raise InterfaceNotFoundException(
                f'[!] Interface "{args.interface}" does not exist.'
            )
        return iface

    @staticmethod
    def ls(_: argparse.Namespace):
        c = Config()
        if c.interfaces:
            print("[i] Showing all interfaces.")
            for k, v in c.interfaces.items():
                print(_FMT_INTERFACES.format(iface=k, port=v.port, host=v.host))
        else:
            print("[i] No interfaces have been defined.")
        return 0

    @staticmethod
    def create(args: argparse.Namespace):
        c = Config()
        # sanitize params
        # TODO(lavajuno): write library for command line forms
        iface_name = str(args.name)
        valid, reason = Input.check_iface(iface_name)
        if not valid:
            print("[!] Interface name is invalid:")
            print(reason)
            return 1
        if c.interfaces.get(iface_name):
            print(
                "[!] An interface with this name already exists! Please choose a different name."
            )
            return 1
        cidr4 = str(args.cidr4)
        if cidr4:
            valid, reason = Input.check_cidr4(cidr4)
            if not valid:
                print("[!] IPv4 CIDR block is invalid:")
                print(reason)
                return 1
        else:
            cidr4 = IP.auto_cidr4()
        cidr6 = str(args.cidr6)
        if cidr6:
            valid, reason = Input.check_cidr6(cidr6)
            if not valid:
                print("[!] IPv4 CIDR block is invalid:")
                print(reason)
                return 1
        else:
            cidr6 = IP.auto_cidr6()
        port = int(args.port)
        valid, reason = Input.check_int(port, min_value=1, max_value=65535)
        if not valid:
            print("[!] Port is invalid:")
            print(reason)
            return 1
        host = str(args.host)
        if not host:
            print("[!] Host is required.")
            return 1
        iface = wireguard.Interface.create(
            vpn_iface=iface_name,
            vpn_cidr4=cidr4,
            vpn_cidr6=cidr6,
            host=host,
            port=port,
        )
        c.interfaces[iface_name] = iface
        c.save()
        print(f'[i] Created interface "{iface_name}".')
        return 0

    @classmethod
    def show(cls, args: argparse.Namespace):
        c = Config()
        iface = cls._get(c, args)
        print(f'[i] Showing interface "{iface.vpn_iface}".')
        print(_FMT_ATTRS.format("Public Key", iface.public_key))
        print(_FMT_ATTRS.format("VPN IPv4 Pool", iface.vpn_cidr4))
        print(_FMT_ATTRS.format("VPN IPv6 Pool", iface.vpn_cidr6))
        print(_FMT_ATTRS.format("NAT", "Enabled" if iface.nat_iface else "Disabled"))
        if iface.nat_iface:
            print(_FMT_ATTRS.format("NAT Interface", iface.nat_iface))
            print(_FMT_ATTRS.format("NAT IPv4 Dests", ", ".join(iface.nat_cidr4)))
            print(_FMT_ATTRS.format("NAT IPv6 Dests", ", ".join(iface.nat_cidr6)))
        return 0

    @classmethod
    def set(cls, args: argparse.Namespace):
        c = Config()
        iface = cls._get(c, args)
        match args.attribute:
            case cls.Attributes.NAME.value:
                raise NotImplementedError
            case cls.Attributes.HOST.value:
                iface.host = args.value
            case cls.Attributes.PORT.value:
                valid, reason = Input.check_int(
                    args.value, min_value=1, max_value=65535
                )
                if not valid:
                    print("[!] Port is invalid:")
                    print(reason)
                    return 1
                iface.port = int(args.value)
            case cls.Attributes.NAT_IFACE.value:
                valid, reason = Input.check_iface(args.value)
                if not valid:
                    print("[!] NAT interface is invalid:")
                    print(reason)
                    return 1
                iface.nat_iface = args.value
            case _:
                print(
                    f"[!] Please specify one of the following attributes: {", ".join(x.value for x in cls.Attributes)}"
                )
                return 1
        print(
            f'[i] Set {args.attribute}="{args.value}" (interface "{args.interface}").'
        )
        c.save()
        return 0

    @classmethod
    def rm(cls, args: argparse.Namespace):
        c = Config()
        _ = cls._get(c, args)  # ignore
        if not args.force:
            print("Are you sure you want to remove this interface?")
            print("This operation is irreversible!")
            print("-> (y/N):")
            if input().lower() != "y":
                print("[!] Operation cancelled by user. No action taken.")
                return 1
        del c.interfaces[args.interface]
        print(f'Removed interface "{args.interface}".')
        c.save()
        return 0

    @classmethod
    def sync(cls, args: argparse.Namespace):
        c = Config()
        iface = cls._get(c, args)
        iface_conf = iface.get_config()
        temp_filename = f"{defaults.CONFIG_DIR}/sync_temp"
        try:
            with open(temp_filename, "w") as f:
                f.write(iface_conf)
        except Exception as e:
            print(f"[!] Could not write temporary file: {str(e)}")
        iface.sync(temp_filename)
        print(f'[i] Synced interface "{args.interface}".')
        return 0

    @classmethod
    def export(cls, args: argparse.Namespace):
        c = Config()
        iface = cls._get(c, args)
        iface_conf = iface.get_config()
        if args.filename:
            try:
                with open(args.filename, "w") as f:
                    f.write(iface_conf)
                print(f'[i] Wrote "{args.filename}"')
            except Exception as e:
                print(f'[!] Could not write "{args.filename}": {str(e)}')
        else:
            print(iface_conf)
            return 1
        return 0

    @classmethod
    def up(cls, args: argparse.Namespace):
        c = Config()
        iface = cls._get(c, args)
        wireguard.CommandLine.service_up(iface.vpn_iface)
        return 0

    @classmethod
    def down(cls, args: argparse.Namespace):
        c = Config()
        iface = cls._get(c, args)
        wireguard.CommandLine.service_down(iface.vpn_iface)
        return 0

    @classmethod
    def reload(cls, args: argparse.Namespace):
        c = Config()
        iface = cls._get(c, args)
        wireguard.CommandLine.service_reload(iface.vpn_iface)
        return 0

    @classmethod
    def nat_create(cls, args: argparse.Namespace):
        if not (args.cidr4 or args.cidr6):
            raise ArgsException(
                "[!] Please specify an IPv4 or IPv6 destination (or both)."
            )
        c = Config()
        iface = cls._get(c, args)
        if args.cidr4 and args.cidr4 not in iface.nat_cidr4:
            valid, reason = Input.check_cidr4(args.cidr4)
            if not valid:
                print("[!] IPv4 CIDR is invalid:")
                print(reason)
                return 1
            iface.nat_cidr4.append(args.cidr4)
        if args.cidr6 not in iface.nat_cidr6:
            valid, reason = Input.check_cidr6(args.cidr6)
            if not valid:
                print("[!] IPv6 CIDR is invalid:")
                print(reason)
                return 1
            iface.nat_cidr6.append(args.cidr6)
        c.save()
        print(f'[i] Created NAT on interface "{args.interface}".')
        return 0

    @classmethod
    def nat_rm(cls, args: argparse.Namespace):
        if not (args.cidr4 or args.cidr6):
            raise ArgsException(
                "[!] Please specify an IPv4 or IPv6 destination (or both)."
            )
        c = Config()
        iface = cls._get(c, args)
        if args.cidr4:
            valid, reason = Input.check_cidr4(args.cidr4)
            if not valid:
                print("[!] IPv4 CIDR is invalid:")
                print(reason)
                return 1
            if args.cidr4 in iface.nat_cidr4:
                iface.nat_cidr4.remove(args.cidr4)
            else:
                raise ArgsException(
                    f'[!] NAT to {args.cidr4} does not exist on interface "{args.interface}".'
                )
        if args.cidr6:
            valid, reason = Input.check_cidr6(args.cidr6)
            if not valid:
                print("[!] IPv6 CIDR is invalid:")
                print(reason)
                return 1
            if args.cidr6 in iface.nat_cidr6:
                iface.nat_cidr6.remove(args.cidr6)
            else:
                raise ArgsException(
                    f'[!] NAT to {args.cidr6} does not exist on interface "{args.interface}".'
                )
            c.save()
        print(f'[i] Removed NAT on interface "{args.interface}".')
        return 0

    @classmethod
    def rekey(cls, args: argparse.Namespace):
        c = Config()
        iface = cls._get(c, args)
        iface.rekey()
        c.save()
        print(f'[i] Rekeyed interface "{args.interface}".')
        print(
            "    Please export its interface and peer configs again to connect using the new keys."
        )
        return 0


class Peer:
    class Attributes(Enum):
        NAME = "name"
        CIDR4 = "cidr4"
        CIDR6 = "cidr6"

    @staticmethod
    def _get(c: Config, args: argparse.Namespace):
        """
        This function intentionally uses args instead of the name of the
        interface+peer to ensure that positional arguments are consistent
        across subcommands.
        """
        iface = c.interfaces.get(args.interface)
        if iface is None:
            raise InterfaceNotFoundException(
                f'[!] Interface "{args.interface}" does not exist.'
            )
        peer = iface.peers.get(args.peer)
        if peer is None:
            raise PeerNotFoundException(f'[!] Peer "{args.peer}" does not exist.')
        return iface, peer

    @staticmethod
    def __next_addr4(interface: wireguard.Interface):
        peer_cidr4 = list(peer.cidr4 for peer in interface.peers.values())
        return IP.next_addr4(interface.vpn_cidr4, peer_cidr4)

    @staticmethod
    def __next_addr6(interface: wireguard.Interface):
        peer_cidr6 = list(peer.cidr6 for peer in interface.peers.values())
        return IP.next_addr6(interface.vpn_cidr6, peer_cidr6)

    @classmethod
    def create(cls, args: argparse.Namespace):
        c = Config()
        if not c.interfaces.get(args.interface):
            print(f'No such interface: "{args.interface}".')
            return 1
        interface = c.interfaces[args.interface]
        # sanitize params
        # TODO(lavajuno): write library for command line forms
        peer_name = str(args.name)
        valid, reason = Input.check_peer_name(peer_name)
        if not valid:
            print("[!] Peer name is invalid:")
            print(reason)
            return 1
        if interface.peers.get(peer_name):
            print(
                "[!] A peer with this name already exists! Please choose a different name."
            )
            return 1
        cidr4 = str(args.cidr4)
        if cidr4:
            valid, reason = Input.check_cidr4(cidr4)
            if not valid:
                print("[!] IPv4 CIDR block is invalid:")
                print(reason)
                return 1
        else:
            cidr4 = cls.__next_addr4(interface)
        cidr6 = str(args.cidr6)
        if cidr6:
            valid, reason = Input.check_cidr6(cidr6)
            if not valid:
                print("[!] IPv6 CIDR block is invalid:")
                print(reason)
                return 1
        else:
            cidr6 = cls.__next_addr6(interface)
        peer = wireguard.Peer.create(name=peer_name, cidr4=cidr4, cidr6=cidr6)
        interface.peers[peer_name] = peer
        c.save()
        print(f'[i] Created peer "{peer_name}" for interface "{args.interface}".')
        return 0

    @classmethod
    def ls(cls, args: argparse.Namespace):
        c = Config()
        iface = c.interfaces.get(args.interface)
        if iface is None:
            print(f'[!] Interface "{args.interface}" does not exist.')
            return 1
        if iface.peers:
            print(f'[i] Showing peers for "{iface.vpn_iface}".')
            for k, v in iface.peers.items():
                print(_FMT_PEERS.format(name=k, cidr4=v.cidr4, cidr6=v.cidr6))
        else:
            print("[i] No peers have been defined for this interface.")
        return 0

    @classmethod
    def show(cls, args: argparse.Namespace):
        c = Config()
        iface = c.interfaces.get(args.interface)
        if iface is None:
            print(f'[!] Interface "{args.interface}" does not exist.')
            return 1
        peer = iface.peers.get(args.peer)
        if peer is None:
            print(f'[!] Peer "{args.peer}" does not exist.')
            return 1
        print(f'[i] Showing peer "{args.peer}" (on interface {args.interface}).')
        print(_FMT_ATTRS.format("Public Key", peer.public_key))
        print(_FMT_ATTRS.format("IPv4 CIDR", peer.cidr4))
        print(_FMT_ATTRS.format("IPv6 CIDR", peer.cidr6))
        return 0

    @classmethod
    def export(cls, args: argparse.Namespace):
        c = Config()
        iface = c.interfaces.get(args.interface)
        if iface is None:
            print(f'[!] Interface "{args.interface}" does not exist.')
            return 1
        peer = iface.peers.get(args.peer)
        if peer is None:
            print(f'[!] Peer "{args.peer}" does not exist.')
            return 1
        peer_conf = peer.get_config(
            vpn_cidr4=iface.vpn_cidr4,
            vpn_cidr6=iface.vpn_cidr6,
            nat_cidr4=iface.nat_cidr4,
            nat_cidr6=iface.nat_cidr6,
            endpoint_public_key=iface.public_key,
            endpoint_host=iface.host,
        )
        if args.filename:
            try:
                with open(args.filename, "w") as f:
                    f.write(peer_conf)
                print(f'[i] Wrote "{args.filename}"')
            except Exception as e:
                print(f'[i] Could not write "{args.filename}": {str(e)}')
        else:
            print(peer_conf)
            return 1
        return 0

    @classmethod
    def set(cls, args: argparse.Namespace):
        c = Config()
        _, peer = cls._get(c, args)
        match args.attribute:
            case cls.Attributes.NAME.value:
                raise NotImplementedError
            case cls.Attributes.CIDR4.value:
                # TODO ensure CIDRs are in network and unused
                valid, reason = Input.check_cidr4(args.value)
                if not valid:
                    print("[!] IPv4 CIDR block is invalid:")
                    print(reason)
                    return 1
                peer.cidr4 = args.value
            case cls.Attributes.CIDR6.value:
                valid, reason = Input.check_cidr6(args.value)
                if not valid:
                    print("[!] IPv6 CIDR block is invalid:")
                    print(reason)
                    return 1
                peer.cidr6 = args.value
            case _:
                print(
                    f"[!] Please specify one of the following attributes: {", ".join(x.value for x in cls.Attributes)}"
                )
                return 1
        print(
            f'[i] Set {args.attribute}="{args.value}" (peer "{args.peer}" on {args.interface}).'
        )
        c.save()
        return 0

    @classmethod
    def rm(cls, args: argparse.Namespace):
        c = Config()
        iface, _ = cls._get(c, args.interface)
        if not args.force:
            print("Are you sure you want to remove this peer?")
            print("This operation is irreversible!")
            print("-> (y/N):")
            if input().lower() != "y":
                print("[!] Operation cancelled by user. No action taken.")
                return 1
        del iface.peers[args.peer]
        print(f'Removed peer "{args.peer}" (on {args.interface}).')
        c.save()
        return 0

    @classmethod
    def rekey(cls, args: argparse.Namespace):
        c = Config()
        _, peer = cls._get(c, args)
        peer.rekey()
        c.save()
        print(f'[i] Rekeyed peer "{args.peer}" (on {args.interface}).')
        print(
            "    Please export the peer's config again to connect using the new keys."
        )
        return 0


class Version:
    @staticmethod
    def display(_: argparse.Namespace):
        print(
            f"{defaults.PROG} v{defaults.VERSION} (config version {defaults.CONFIG_VERSION})"
        )
        return 0


@staticmethod
def get_parser():
    # root
    root = argparse.ArgumentParser(defaults.PROG)
    root_sub = root.add_subparsers(title="subcommands", required=True)

    # iface.*
    iface = root_sub.add_parser("iface", help="Manage interfaces")
    iface_sub = iface.add_subparsers(title="subcommands", required=True)

    # iface.ls
    interface_ls = iface_sub.add_parser("ls", help="List all interfaces")
    interface_ls.set_defaults(func=Iface.ls)

    # iface.create
    iface_create = iface_sub.add_parser("create", help="Create an interface")
    iface_create.set_defaults(func=Iface.create)
    iface_create.add_argument("name", type=str, default="")
    iface_create.add_argument("--cidr4", type=str, default="")
    iface_create.add_argument("--cidr6", type=str, default="")
    iface_create.add_argument("--nat_iface", type=str, default="")
    iface_create.add_argument("--host", type=str, required=True)
    iface_create.add_argument("--port", type=int, required=True)

    # iface.show
    iface_show = iface_sub.add_parser("show", help="Show details for an interface")
    iface_show.set_defaults(func=Iface.show)
    iface_show.add_argument("interface", type=str)

    # iface.set
    iface_set = iface_sub.add_parser("set", help="Set parameters for an interface")
    iface_set.set_defaults(func=Iface.set)
    iface_set.add_argument("interface", type=str)
    iface_set.add_argument("attribute", type=str)
    iface_set.add_argument("value", type=str)

    # iface.remove
    iface_rm = iface_sub.add_parser("rm", help="Remove an interface")
    iface_rm.set_defaults(func=Iface.rm)
    iface_rm.add_argument("interface", type=str)
    iface_rm.add_argument("--force", action="store_true")

    # iface.export
    iface_export = iface_sub.add_parser(
        "export", help="Export config file for an interface"
    )
    iface_export.set_defaults(func=Iface.export)
    iface_export.add_argument("interface", type=str)
    iface_export.add_argument("-f", "--filename", type=str)

    # iface.up
    iface_up = iface_sub.add_parser(
        "up", help="Bring an interface up and enable its systemd target"
    )
    iface_up.set_defaults(func=Iface.up)
    iface_up.add_argument("interface", type=str)

    # iface.down
    iface_down = iface_sub.add_parser(
        "down", help="Bring an interface down and disable its systemd target"
    )
    iface_down.set_defaults(func=Iface.down)
    iface_down.add_argument("interface", type=str)

    # iface.reload
    iface_reload = iface_sub.add_parser(
        "reload", help="Tell systemd to reload an interface's config"
    )
    iface_reload.set_defaults(func=Iface.reload)
    iface_reload.add_argument("interface", type=str)

    # iface.rekey
    iface_rekey = iface_sub.add_parser(
        "rekey", help="Generate new keys for an interface and its peers"
    )
    iface_rekey.set_defaults(func=Iface.rekey)
    iface_rekey.add_argument("interface", type=str)

    # iface.sync
    iface_sync = iface_sub.add_parser(
        "sync", help="Update an interface's config in /etc/wireguard"
    )
    iface_sync.set_defaults(func=Iface.sync)
    iface_sync.add_argument("interface", type=str)

    # peer.*
    peer = root_sub.add_parser("peer", help="Manage peers")
    peer_sub = peer.add_subparsers()

    # peer.ls
    peer_ls = peer_sub.add_parser("ls", help="Show peers defined for an interface")
    peer_ls.set_defaults(func=Peer.ls)
    peer_ls.add_argument("interface", type=str)

    # peer.create
    peer_create = peer_sub.add_parser(
        "create", help="Create a new peer for the given interface"
    )
    peer_create.set_defaults(func=Peer.create)
    peer_create.add_argument("interface", type=str)
    peer_create.add_argument("name", type=str, default="")
    peer_create.add_argument("--cidr4", type=str, default="")
    peer_create.add_argument("--cidr6", type=str, default="")

    # peer.show
    peer_show = peer_sub.add_parser("show", help="Show details for a peer")
    peer_show.set_defaults(func=Peer.show)
    peer_show.add_argument("interface", type=str)
    peer_show.add_argument("peer", type=str)

    # peer.export
    peer_export = peer_sub.add_parser("export", help="Export config file for a peer")
    peer_export.set_defaults(func=Peer.export)
    peer_export.add_argument("interface", type=str)
    peer_export.add_argument("peer", type=str)
    peer_export.add_argument("-f", "--filename", type=str)

    # peer.set
    peer_set = peer_sub.add_parser("set")
    peer_set.set_defaults(func=Peer.set)
    peer_set.add_argument("interface", type=str)
    peer_set.add_argument("peer", type=str)
    peer_set.add_argument("attribute", type=str)
    peer_set.add_argument("value", type=str)

    # peer.rm
    peer_rm = peer_sub.add_parser("rm")
    peer_rm.set_defaults(func=Peer.rm)
    peer_rm.add_argument("interface", type=str)
    peer_rm.add_argument("peer", type=str)
    peer_rm.add_argument("--force", action="store_true")

    # peer.rekey
    peer_rekey = peer_sub.add_parser("rekey")
    peer_rekey.set_defaults(func=Peer.rekey)
    peer_rekey.add_argument("interface", type=str)
    peer_rekey.add_argument("peer", type=str)

    # nat.*
    nat = root_sub.add_parser("nat", help="Manage NATs")
    nat_sub = nat.add_subparsers(title="subcommands", required=True)

    # nat.create
    nat_create = nat_sub.add_parser("create", help="Create a NAT")
    nat_create.set_defaults(func=Iface.nat_create)
    nat_create.add_argument("interface", type=str)
    nat_create.add_argument("--cidr4", type=str, default="")
    nat_create.add_argument("--cidr6", type=str, default="")

    # nat.rm
    nat_rm = nat_sub.add_parser("rm", help="Remove a NAT")
    nat_rm.set_defaults(func=Iface.nat_rm)
    nat_rm.add_argument("interface", type=str)
    nat_rm.add_argument("--cidr4", type=str, default="")
    nat_rm.add_argument("--cidr6", type=str, default="")

    # version
    version = root_sub.add_parser("version", help="Show version information")
    version.set_defaults(func=Version.display)

    return root  # parser


def entrypoint():
    parser = get_parser()
    args = parser.parse_args(sys.argv[1:])
    try:
        status = int(args.func(args))
        return status
    except ExitException as e:
        print(str(e))
        return 1



import json
import subprocess

from wgup import defaults
from wgup.util import IP

CONFIG_FW_VPN_FWD = """
# Firewall: Allow traffic flow within VPN interface
PostUp = iptables -I FORWARD -i {vpn_iface} -o {vpn_iface} -j ACCEPT
PostUp = ip6tables -I FORWARD -i {vpn_iface} -o {vpn_iface} -j ACCEPT

PreDown = iptables -D FORWARD -i {vpn_iface} -o {vpn_iface} -j ACCEPT
PreDown = ip6tables -D FORWARD -i {vpn_iface} -o {vpn_iface} -j ACCEPT
"""

CONFIG_FW_NAT_FLOW = """
# Firewall: Allow traffic flow between VPN interface and NAT interface
PostUp = iptables -I FORWARD -i {vpn_iface} -o {nat_iface} -j ACCEPT
PostUp = ip6tables -I FORWARD -i {vpn_iface} -o {nat_iface} -j ACCEPT
PostUp = iptables -I FORWARD -i {nat_iface} -o {vpn_iface} -j ACCEPT
PostUp = ip6tables -I FORWARD -i {nat_iface} -o {vpn_iface} -j ACCEPT

PreDown = iptables -D FORWARD -i {vpn_iface} -o {nat_iface} -j ACCEPT
PreDown = ip6tables -D FORWARD -i {vpn_iface} -o {nat_iface} -j ACCEPT
PreDown = iptables -D FORWARD -i {nat_iface} -o {vpn_iface} -j ACCEPT
PreDown = ip6tables -D FORWARD -i {nat_iface} -o {vpn_iface} -j ACCEPT
"""

CONFIG_FW_NAT_DEST = """
# Firewall: NAT the following traffic coming from the VPN
PostUp = iptables -t nat -I POSTROUTING -o {nat_iface} -d {nat_cidr4} -j MASQUERADE
PostUp = ip6tables -t nat -I POSTROUTING -o {nat_iface} -d {nat_cidr6} -j MASQUERADE

PreDown = iptables -t nat -D POSTROUTING -o {nat_iface} -d {nat_cidr4} -j MASQUERADE
PreDown = ip6tables -t nat -D POSTROUTING -o {nat_iface} -d {nat_cidr6} -j MASQUERADE
"""

CONFIG_PEER_HEADER = """
# Peer "{name}"
[Interface]
PrivateKey = {private_key}
Address = {cidr4}
Address = {cidr6}
"""

CONFIG_PEER_ENDPOINT = """
[Peer]
PublicKey = {public_key}
PresharedKey = {preshared_key}
AllowedIPs = {cidr4}
AllowedIPs = {cidr6}
Endpoint = {endpoint}
PersistentKeepalive = 10  # Keeps peers behind NATs accessible
"""

CONFIG_INTERFACE_HEADER = """
[Interface]
PrivateKey = {private_key}
Address = {cidr4}
Address = {cidr6}
ListenPort = {port}
"""

CONFIG_INTERFACE_PEER = """
# Peer "{name}"
[Peer]
PublicKey = {public_key}
PresharedKey = {preshared_key}
AllowedIPs = {cidr4}
AllowedIPs = {cidr6}
"""


class CommandLine:
    @staticmethod
    def generate_private_key() -> str:
        result = subprocess.run(["wg", "genkey"], stdout=subprocess.PIPE)
        result.check_returncode()
        return result.stdout.decode("utf-8").strip()

    @staticmethod
    def generate_public_key(private_key: str) -> str:
        result = subprocess.run(
            ["wg", "pubkey"], stdout=subprocess.PIPE, input=private_key.encode("utf-8")
        )
        result.check_returncode()
        return result.stdout.decode("utf-8").strip()

    @staticmethod
    def generate_preshared_key() -> str:
        result = subprocess.run(["wg", "genpsk"], stdout=subprocess.PIPE)
        result.check_returncode()
        return result.stdout.decode("utf-8").strip()

    @staticmethod
    def service_up(if_name: str):
        subprocess.run(
            ["sudo", "systemctl", "enable", f"wg-quick@{if_name}"],
        ).check_returncode()
        subprocess.run(
            ["sudo", "systemctl", "restart", f"wg-quick@{if_name}"],
        ).check_returncode()

    @staticmethod
    def service_down(if_name: str):
        subprocess.run(
            ["sudo", "systemctl", "disable", f"wg-quick@{if_name}"],
        ).check_returncode()
        subprocess.run(
            ["sudo", "systemctl", "stop", f"wg-quick@{if_name}"],
        ).check_returncode()

    @staticmethod
    def service_reload(if_name: str):
        subprocess.run(
            ["sudo", "systemctl", "reload", f"wg-quick@{if_name}"],
        ).check_returncode()

    @staticmethod
    def copy_config(if_name: str, source_file: str):
        result = subprocess.run(
            ["sudo", "mv", source_file, f"/etc/wireguard/{if_name}.conf"],
        )
        result.check_returncode()


class Peer:
    def __init__(
        self,
        *,
        name: str,
        private_key: str,
        public_key: str,
        preshared_key: str,
        cidr4: str,
        cidr6: str,
    ):
        self.name = name
        self.private_key = private_key
        self.public_key = public_key
        self.preshared_key = preshared_key
        self.cidr4 = cidr4
        self.cidr6 = cidr6

    @classmethod
    def create(cls, *, name: str, cidr4: str, cidr6: str):
        private_key = CommandLine.generate_private_key()
        public_key = CommandLine.generate_public_key(private_key)
        preshared_key = CommandLine.generate_preshared_key()
        return cls(
            name=name,
            private_key=private_key,
            public_key=public_key,
            preshared_key=preshared_key,
            cidr4=cidr4,
            cidr6=cidr6,
        )

    def rekey(self):
        self.private_key = CommandLine.generate_private_key()
        self.public_key = CommandLine.generate_public_key(self.private_key)
        self.preshared_key = CommandLine.generate_preshared_key()

    def __get_peer_header(self) -> str:
        return CONFIG_PEER_HEADER.format(
            name=self.name,
            private_key=self.private_key,
            cidr4=self.cidr4,
            cidr6=self.cidr6,
        )

    def __get_peer_endpoint(
        self,
        public_key: str,
        vpn_cidr4: str,
        vpn_cidr6: str,
        nat_cidr4: list[str],
        nat_cidr6: list[str],
        endpoint: str,
    ) -> str:
        return CONFIG_PEER_ENDPOINT.format(
            public_key=public_key,
            preshared_key=self.preshared_key,
            cidr4=",".join([vpn_cidr4, *nat_cidr4]),
            cidr6=",".join([vpn_cidr6, *nat_cidr6]),
            endpoint=endpoint,
        )

    def get_config(
        self,
        vpn_cidr4: str,
        vpn_cidr6: str,
        nat_cidr4: list[str],
        nat_cidr6: list[str],
        endpoint_public_key: str,
        endpoint_host: str,
    ):
        return "# Generated by {} v{}\n{}{}".format(
            defaults.PROG,
            defaults.VERSION,
            self.__get_peer_header(),
            self.__get_peer_endpoint(
                endpoint_public_key,
                vpn_cidr4,
                vpn_cidr6,
                nat_cidr4,
                nat_cidr6,
                endpoint_host,
            ),
        )

    def to_json(self):
        return {
            "name": self.name,
            "private_key": self.private_key,
            "public_key": self.public_key,
            "preshared_key": self.preshared_key,
            "cidr4": self.cidr4,
            "cidr6": self.cidr6,
        }

    @classmethod
    def from_json(cls, data: dict):
        return cls(
            name=data["name"],
            private_key=data["private_key"],
            public_key=data["public_key"],
            preshared_key=data["preshared_key"],
            cidr4=data["cidr4"],
            cidr6=data["cidr6"],
        )


class Interface:
    def __init__(
        self,
        *,
        private_key: str,
        public_key: str,
        vpn_iface: str,
        vpn_cidr4: str,
        vpn_cidr6: str,
        addr4: str,
        addr6: str,
        host: str,
        port: int,
        nat_iface: str | None = None,
        nat_cidr4: list[str] | None = None,
        nat_cidr6: list[str] | None = None,
        dns: list[str] | None = None,
        peers: dict[str, Peer] | None = None,
    ):
        self.private_key = private_key
        self.public_key = public_key
        self.vpn_iface = vpn_iface
        self.vpn_cidr4 = vpn_cidr4
        self.vpn_cidr6 = vpn_cidr6
        self.addr4 = addr4
        self.addr6 = addr6
        self.host = host
        self.port = port
        if nat_iface is None:
            nat_iface = ""
        self.nat_iface = nat_iface
        if nat_cidr4 is None:
            nat_cidr4 = []
        self.nat_cidr4 = nat_cidr4
        if nat_cidr6 is None:
            nat_cidr6 = []
        self.nat_cidr6 = nat_cidr6
        if dns is None:
            dns = []
        self.dns = dns
        if peers is None:
            peers = {}
        self.peers = peers

    @classmethod
    def create(
        cls,
        *,
        vpn_iface: str,
        vpn_cidr4: str,
        vpn_cidr6: str,
        host: str,
        port: int,
    ):
        private_key = CommandLine.generate_private_key()
        public_key = CommandLine.generate_public_key(private_key)
        return cls(
            private_key=private_key,
            public_key=public_key,
            vpn_iface=vpn_iface,
            vpn_cidr4=vpn_cidr4,
            vpn_cidr6=vpn_cidr6,
            addr4=IP.server_addr4(vpn_cidr4),
            addr6=IP.server_addr6(vpn_cidr6),
            host=host,
            port=port,
        )

    def __get_fw_vpn_fwd(self) -> str:
        return CONFIG_FW_VPN_FWD.format(vpn_iface=self.vpn_iface)

    def __get_fw_nat_flow(self) -> str:
        return CONFIG_FW_NAT_FLOW.format(
            vpn_iface=self.vpn_iface, nat_iface=self.nat_iface
        )

    def __get_fw_nat_dest(self) -> str:
        return CONFIG_FW_NAT_DEST.format(
            nat_iface=self.nat_iface,
            nat_cidr4=",".join(self.nat_cidr4),
            nat_cidr6=",".join(self.nat_cidr6),
        )

    def __get_network_header(self) -> str:
        return CONFIG_INTERFACE_HEADER.format(
            private_key=self.private_key,
            cidr4=self.addr4,
            cidr6=self.addr6,
            port=self.port,
        )

    def __get_peers_config(self):
        return "".join(
            CONFIG_INTERFACE_PEER.format(
                name=peer.name,
                public_key=peer.public_key,
                preshared_key=peer.preshared_key,
                cidr4=peer.cidr4,
                cidr6=peer.cidr6,
            )
            for peer in self.peers.values()
        )

    def __get_nat_config(self) -> str:
        if not self.nat_iface:
            return ""
        return "{}{}".format(
            self.__get_fw_nat_flow(),
            self.__get_fw_nat_dest(),
        )

    def get_config(self) -> str:
        return "# Generated by {} v{}\n{}{}{}{}".format(
            defaults.PROG,
            defaults.VERSION,
            self.__get_network_header(),
            self.__get_fw_vpn_fwd(),
            self.__get_nat_config(),
            self.__get_peers_config(),
        )

    def sync(self, source_file: str):
        CommandLine.copy_config(self.vpn_iface, source_file)

    def rekey(self):
        self.private_key = CommandLine.generate_private_key()
        self.public_key = CommandLine.generate_public_key(self.private_key)
        for peer in self.peers.values():
            peer.rekey()

    def to_json(self):
        return {
            "private_key": self.private_key,
            "public_key": self.public_key,
            "vpn_iface": self.vpn_iface,
            "vpn_cidr4": self.vpn_cidr4,
            "vpn_cidr6": self.vpn_cidr6,
            "addr4": self.addr4,
            "addr6": self.addr6,
            "host": self.host,
            "port": self.port,
            "nat_iface": self.nat_iface,
            "nat_cidr4": self.nat_cidr4,
            "nat_cidr6": self.nat_cidr6,
            "peers": list(p[1].to_json() for p in sorted(self.peers.items())),
        }

    @classmethod
    def from_json(cls, data: dict):
        peers: dict[str, Peer] = {}
        for p in data["peers"]:
            peer = Peer.from_json(p)
            peers[peer.name] = peer
        return cls(
            private_key=data["private_key"],
            public_key=data["public_key"],
            vpn_iface=data["vpn_iface"],
            vpn_cidr4=data["vpn_cidr4"],
            vpn_cidr6=data["vpn_cidr6"],
            addr4=data["addr4"],
            addr6=data["addr6"],
            host=data["host"],
            port=int(data["port"]),
            nat_iface=data["nat_iface"],
            nat_cidr4=data["nat_cidr4"],
            nat_cidr6=data["nat_cidr6"],
            peers=peers,
        )



import json
import logging
import os

from wgup import defaults
from wgup.util import ConfigVersionException
from wgup.wireguard import Interface

_CONFIG_INTERFACES = f"{defaults.CONFIG_DIR}/interfaces.json"

_logger = logging.getLogger(defaults.PROG)


class Config:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(Config, cls).__new__(cls)
            cls._instance._setup()
        return cls._instance

    def _setup(self):
        self.interfaces: dict[str, Interface] = {}
        os.makedirs(defaults.CONFIG_DIR, exist_ok=True)
        self.load()

    def load(self):
        if not os.path.exists(_CONFIG_INTERFACES):
            return
        with open(_CONFIG_INTERFACES, "rb") as f:
            networks_json = json.loads(f.read())
        if networks_json["version"] != defaults.CONFIG_VERSION:
            raise ConfigVersionException("[!] Incompatible config version.")
        for n in networks_json["interfaces"]:
            interface = Interface.from_json(n)
            self.interfaces[interface.vpn_iface] = interface

    def save(self):
        interfaces_json = {
            "version": defaults.CONFIG_VERSION,
            "interfaces": list(i[1].to_json() for i in sorted(self.interfaces.items())),
        }
        with open(_CONFIG_INTERFACES, "w") as f:
            f.write(json.dumps(interfaces_json, indent=4))
        _logger.debug("Saved configuration.")



import json
from unittest import TestCase

from wgup import wireguard
from wgup.util import IP


class TestWireguard(TestCase):
    def setUp(self):
        self.iface = wireguard.Interface.create(
            vpn_iface="wg0",
            vpn_cidr4=IP.auto_cidr4(),
            vpn_cidr6=IP.auto_cidr6(),
            host="example.com",
            port=12345,
        )
        self.peer0 = wireguard.Peer.create(
            name="peer0",
            cidr4=IP.next_addr4(self.iface.vpn_cidr4, []),
            cidr6=IP.next_addr6(self.iface.vpn_cidr6, []),
        )

    def test_iface_json(self):
        orig_iface_json = json.dumps(self.iface.to_json())
        loaded_iface = wireguard.Interface.from_json(json.loads(orig_iface_json))
        loaded_iface_json = json.dumps(loaded_iface.to_json())
        self.assertEqual(orig_iface_json, loaded_iface_json)
